#include <sys/param.h>
#include <sys/libkern.h>

#include <machine/cons.h>
#include <machine/prom.h>
#include <machine/stdarg.h>	/* should be <stdarg.h> in a selfhosted
				   environment */

static handle_t prom_root, prom_stdout, prom_stdin, prom_chosen;

void
prom_init()
{

	/* XXX check for retval of -1 */
	prom_root = prom_peer(0);
	prom_chosen = prom_finddevice("/chosen");
	prom_getprop(prom_chosen, "stdout", (void *)&prom_stdout,
		      sizeof prom_stdout);
	prom_getprop(prom_chosen, "stdin", (void *)&prom_stdin,
		      sizeof prom_stdin);
	prom_printf("prom_init: prom_stdout = 0x%0lx\n", prom_stdout);
	prom_printf("prom_init: prom_root = 0x%0lx\n", prom_root);
}

/* KCH: next two borrowed from kern/subr_prf.c */
/*
 * Put a number (base <= 16) in a buffer in reverse order; return an
 * optional length and a pointer to the NULL terminated (preceded?)
 * buffer.
 */
static char *
ksprintn(ul, base, lenp)
	register u_long ul;
	register int base, *lenp;
{					/* A long in base 8, plus NULL. */
	static char buf[sizeof(long) * NBBY / 3 + 2];
	register char *p;

	p = buf;
	do {
		*++p = hex2ascii(ul % base);
	} while (ul /= base);
	if (lenp)
		*lenp = p - buf;
	return (p);
}

/*
 * Scaled down version of printf(3).
 *
 * Two additional formats:
 *
 * The format %b is supported to decode error registers.
 * Its usage is:
 *
 *	printf("reg=%b\n", regval, "<base><arg>*");
 *
 * where <base> is the output base expressed as a control character, e.g.
 * \10 gives octal; \20 gives hex.  Each arg is a sequence of characters,
 * the first of which gives the bit number to be inspected (origin 1), and
 * the next characters (up to a control character, i.e. a character <= 32),
 * give the name of the register.  Thus:
 *
 *	kvprintf("reg=%b\n", 3, "\10\2BITTWO\1BITONE\n");
 *
 * would produce output:
 *
 *	reg=3<BITTWO,BITONE>
 *
 * XXX:  %D  -- Hexdump, takes pointer and separator string:
 *		("%6D", ptr, ":")   -> XX:XX:XX:XX:XX:XX
 *		("%*D", len, ptr, " " -> XX XX XX XX ...
 */
int
prom_printf(char const *fmt, ...)
{
#define PCHAR(c) prom_putchar(c)
	static char *ksprintn(u_long, int, int*);
	static int radix=10;
	va_list ap;
	char *p, *q;
	u_char *up;
	int ch, n;
	u_long ul;
	int base, lflag, tmp, width, ladjust, sharpflag, neg, sign, dot;
	int dwidth;
	char padc;
	int retval = 0;

	va_start(ap, fmt);

	if (fmt == NULL)
		fmt = "(fmt null)\n";

	if (radix < 2 || radix > 36)
		radix = 10;

	for (;;) {
		padc = ' ';
		width = 0;
		while ((ch = (u_char)*fmt++) != '%') {
			if (ch == '\0') {
				va_end(ap);
				return retval;
			}
			PCHAR(ch);
		}
		lflag = 0; ladjust = 0; sharpflag = 0; neg = 0;
		sign = 0; dot = 0; dwidth = 0;
reswitch:	switch (ch = (u_char)*fmt++) {
		case '.':
			dot = 1;
			goto reswitch;
		case '#':
			sharpflag = 1;
			goto reswitch;
		case '+':
			sign = 1;
			goto reswitch;
		case '-':
			ladjust = 1;
			goto reswitch;
		case '%':
			PCHAR(ch);
			break;
		case '*':
			if (!dot) {
				width = va_arg(ap, int);
				if (width < 0) {
					ladjust = !ladjust;
					width = -width;
				}
			} else {
				dwidth = va_arg(ap, int);
			}
			goto reswitch;
		case '0':
			if (!dot) {
				padc = '0';
				goto reswitch;
			}
		case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
				for (n = 0;; ++fmt) {
					n = n * 10 + ch - '0';
					ch = *fmt;
					if (ch < '0' || ch > '9')
						break;
				}
			if (dot)
				dwidth = n;
			else
				width = n;
			goto reswitch;
		case 'b':
			ul = va_arg(ap, int);
			p = va_arg(ap, char *);
			for (q = ksprintn(ul, *p++, NULL); *q;)
				PCHAR(*q--);

			if (!ul)
				break;

			for (tmp = 0; *p;) {
				n = *p++;
				if (ul & (1 << (n - 1))) {
					PCHAR(tmp ? ',' : '<');
					for (; (n = *p) > ' '; ++p)
						PCHAR(n);
					tmp = 1;
				} else
					for (; *p > ' '; ++p)
						continue;
			}
			if (tmp)
				PCHAR('>');
			break;
		case 'c':
			PCHAR(va_arg(ap, int));
			break;
		case 'D':
			up = va_arg(ap, u_char *);
			p = va_arg(ap, char *);
			if (!width)
				width = 16;
			while(width--) {
				PCHAR(hex2ascii(*up >> 4));
				PCHAR(hex2ascii(*up & 0x0f));
				up++;
				if (width)
					for (q=p;*q;q++)
						PCHAR(*q);
			}
			break;
		case 'd':
			ul = lflag ? va_arg(ap, long) : va_arg(ap, int);
			sign = 1;
			base = 10;
			goto number;
		case 'l':
			lflag = 1;
			goto reswitch;
		case 'n':
			ul = lflag ? va_arg(ap, u_long) : va_arg(ap, u_int);
			base = radix;
			goto number;
		case 'o':
			ul = lflag ? va_arg(ap, u_long) : va_arg(ap, u_int);
			base = 8;
			goto number;
		case 'p':
			ul = (u_long)va_arg(ap, void *);
			base = 16;
			sharpflag = 1;
			goto number;
		case 's':
			p = va_arg(ap, char *);
			if (p == NULL)
				p = "(null)";
			if (!dot)
				n = strlen (p);
			else
				for (n = 0; n < dwidth && p[n]; n++)
					continue;

			width -= n;

			if (!ladjust && width > 0)
				while (width--)
					PCHAR(padc);
			while (n--)
				PCHAR(*p++);
			if (ladjust && width > 0)
				while (width--)
					PCHAR(padc);
			break;
		case 'u':
			ul = lflag ? va_arg(ap, u_long) : va_arg(ap, u_int);
			base = 10;
			goto number;
		case 'x':
			ul = lflag ? va_arg(ap, u_long) : va_arg(ap, u_int);
			base = 16;
number:			if (sign && (long)ul < 0L) {
				neg = 1;
				ul = -(long)ul;
			}
			p = ksprintn(ul, base, &tmp);
			if (sharpflag && ul != 0) {
				if (base == 8)
					tmp++;
				else if (base == 16)
					tmp += 2;
			}
			if (neg)
				tmp++;

			if (!ladjust && width && (width -= tmp) > 0)
				while (width--)
					PCHAR(padc);
			if (neg)
				PCHAR('-');
			if (sharpflag && ul != 0) {
				if (base == 8) {
					PCHAR('0');
				} else if (base == 16) {
					PCHAR('0');
					PCHAR('x');
				}
			}

			while (*p)
				PCHAR(*p--);

			if (ladjust && width && (width -= tmp) > 0)
				while (width--)
					PCHAR(padc);

			break;
		default:
			PCHAR('%');
			if (lflag)
				PCHAR('l');
			PCHAR(ch);
			break;
		}
	}
#undef PCHAR
}

/* -1 on failure, 0 on success */
int
prom_puts(const char *str)
{

	while (*str != '\0')
		if (prom_putchar((int)*str++) != 0)
			return -1;

	return 0;
}

/* 0 on success */
int
prom_putchar(int c)
{
	int i;

	i = (int)prom_write(prom_stdout, (void *)&c, sizeof c);
	if (c == '\n') {
		c = '\r';
		(void)prom_write(prom_stdout, (void *)&c, sizeof c);
	}
	return i;
}

cell_t
prom_write(handle_t ihandle, const void *addr, size_t len)
{
	cell_t args[] = {(cell_t)"write", 3, 1, 0, 0, 0, 0};

	args[P1275_FIRST_IN + 0] = (cell_t)ihandle;
	args[P1275_FIRST_IN + 1] = (cell_t)addr;
	args[P1275_FIRST_IN + 2] = (cell_t)len;
	prom_cmd(args);

	return args[P1275_FIRST_IN + 3];
}

cell_t
prom_getprop(handle_t device, const char *prop, void *buf, size_t sz)
{
	cell_t args[] = {(cell_t)"getprop", 4, 1, 0, 0, 0, 0, 0};

	args[P1275_FIRST_IN + 0] = (cell_t)device;
	args[P1275_FIRST_IN + 1] = (cell_t)prop;
	args[P1275_FIRST_IN + 2] = (cell_t)buf;
	args[P1275_FIRST_IN + 3] = (cell_t)sz;
	prom_cmd(args);

	return args[P1275_FIRST_IN + 4];
}

void *
prom_getpropaddr(handle_t device, const char *prop)
{
	unsigned long rv;

	if (prom_getprop(device, prop, (void *)&rv, sizeof(rv)) == -1UL)
		return 0;

	rv >>= 32;
	return (void *)rv;
}

handle_t
prom_peer(handle_t device)
{
	cell_t args[] = {(cell_t)"peer", 1, 1, 0, 0};

	args[P1275_FIRST_IN + 0] = (cell_t)device;
	prom_cmd(args);

	return (handle_t)args[P1275_FIRST_IN + 1];
}

handle_t
prom_finddevice(const char *devspec)
{
	cell_t args[] = {(cell_t)"finddevice", 1, 1, 0, 0};

	args[P1275_FIRST_IN + 0] = (cell_t)devspec;
	prom_cmd(args);

	return (handle_t)args[P1275_FIRST_IN + 1];
}

void
prom_exit(void)
{
	cell_t args[] = {(cell_t)"exit", 0, 0};

	for (;;)
		prom_cmd(args);
}

void
prom_enter(void)
{
	cell_t args[] = {(cell_t)"enter", 0, 0};

	prom_cmd(args);
}

void
prom_set_trap_table(unsigned long *tba)
{
	cell_t args[] = {(cell_t)"SUNW,set-trap-table", 1, 0,
			 (unsigned long)tba};

	prom_cmd(args);
}

void
promcnprobe(struct consdev *cp)
{
	cp->cn_pri = CN_NORMAL;
}

void
promcninit(struct consdev *cp)
{
}

void
promcnputc(dev_t dev, int c)
{
	prom_putchar((char)c);
}
